/// Finds links and can check if they're dead.
///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.
import 'dart:async';
import 'dart:io';

import 'package:markdown/markdown.dart';

///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.

///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.

///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.

///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.

///
/// This library can test any type of file, from markdown to Java.
///
/// Links must be specified with the HTTP or HTTPS protocol to be found. For
/// example, https://google.com.
///
/// A "bad link" is a link whose status code retrieved upon making a request to
/// it isn't in the range of 200-299. A list of all HTTP status codes are here:
/// https://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection. The
/// only extra status code used in this library is 0, which indicates that a
/// server couldn't be contacted because the website itself doesn't exist.

/// Checks a directory for bad links.
///
/// If [directory] isn't specified, the project's root directory is used. Set
/// [recursive] to false if you don't want to check subdirectories.
/// [blacklistedDirectories] and [blacklistedFiles] won't be searched.
Stream<BadLink> getBadLinksInDirectory(
    {final String directory,
    final bool recursive = true,
    final List<BlacklistedDirectory> blacklistedDirectories,
    final List<String> blacklistedFiles}) async* {
  await for (final data in getLinksInDirectory(
      directory: directory,
      recursive: recursive,
      blacklistedDirectories: blacklistedDirectories,
      blacklistedFiles: blacklistedFiles)) {
    final statusCode = await getLinkStatus(data.link);
    if (statusCode != null) {
      yield BadLink.fromLinkLocation(data, statusCode);
    }
  }
}

class BlacklistedDirectory {
  /// The path to this blacklisted directory.
  final String path;

  /// If true, subdirectories will also be blacklisted.
  final bool recursive;

  // If true, symbolic links will be followed.
  final bool followLinks;

  BlacklistedDirectory(this.path,
      {this.recursive = true, this.followLinks = true});
}

/// The path to the file having the bad link, the bad link, and its status code.
class BadLink {
  final String path;
  final String link;
  final int statusCode;

  BadLink(this.path, this.link, this.statusCode);

  BadLink.fromLinkLocation(LinkLocation location, this.statusCode)
      : path = location.path,
        link = location.link;

  BadLink.fromBadLinkStatus(this.path, BadLinkStatus status)
      : link = status.link,
        statusCode = status.statusCode;

  @override
  String toString() => 'Path: $path\nLink: $link\nStatus code: $statusCode';
}

/// Yields the links found the in a directory.
///
/// If [directory] isn't specified, the project's root directory is used. Set
/// [recursive] to false if you don't want to search subdirectories.
/// [blacklistedDirectories] and [blacklistedFiles] won't be searched.
Stream<LinkLocation> getLinksInDirectory(
    {final String directory,
    final bool recursive = true,
    final List<BlacklistedDirectory> blacklistedDirectories,
    List<String> blacklistedFiles}) async* {
  final dir = directory == null ? Directory.current : Directory(directory);
  if (blacklistedFiles == null) {
    blacklistedFiles = [];
  }

  if (blacklistedDirectories != null) {
    for (final directory in blacklistedDirectories) {
      await for (final entity in Directory(directory.path).list(
          recursive: directory.recursive, followLinks: directory.followLinks)) {
        if (entity is! File) {
          continue;
        }

        blacklistedFiles.add(entity.path);
      }
    }
  }

  await for (final entity in dir.list(recursive: recursive)) {
    if (entity is! File) {
      continue;
    }
    if (blacklistedFiles.contains(entity.path)) {
      blacklistedFiles.remove(entity.path);
      continue;
    }

    await for (final link in getLinksInFile(entity.path)) {
      yield LinkLocation(entity.path, link);
    }
  }
}

/// The path to the file containing the link, and the link.
class LinkLocation {
  final String path;
  final String link;

  LinkLocation(this.path, this.link);

  @override
  String toString() => '$path: $link';
}

/// Yields bad links found in the file located at [path].
Stream<BadLinkStatus> getBadLinksInFile(final String path) async* {
  await for (final link in getLinksInFile(path)) {
    final statusCode = await getLinkStatus(link);
    if (statusCode != null) {
      yield BadLinkStatus(link, statusCode);
    }
  }
}

/// The link a request was made to, and the status code returned.
class BadLinkStatus {
  final String link;
  final int statusCode;

  BadLinkStatus(this.link, this.statusCode);

  @override
  String toString() => '$link: $statusCode';
}

/// Yields links found in the file at [path].
Stream<String> getLinksInFile(final String path) async* {
  String content;
  try {
    content = await File(path).readAsString();
  } on FileSystemException {
    return; // The file isn't readable. For example, it's a binary or directory.
  }
  if (_isMarkdown(path)) {
    content = markdownToHtml(content);
  }

  final matches = RegExp(
    r"https?://[A-Za-z0-9-._~:/?#[\]@!$&'()*+,;=`%]+",
    caseSensitive: false,
  ).allMatches(content);
  for (final match in matches) {
    yield match.input.substring(match.start, match.end);
  }
}

/// Returns the status code received after making a request to [link].
///
/// 0 will be returned if the website doesn't exist.
/// null will be returned if [link] isn't bad.
Future<int> getLinkStatus(final String link) async {
  HttpClientRequest request;
  try {
    request = await HttpClient().getUrl(Uri.parse(link));
  } on SocketException {
    return 0; // Website doesn't exist.
  }
  final response = await request.close();
  return RegExp(r'2\d\d').hasMatch(response.statusCode.toString())
      ? null
      : response.statusCode;
}

/// Returns true if the file at [path] is of type markdown.
bool _isMarkdown(final String path) {
  for (final extension in [
    '.markdown',
    '.mdown',
    '.mkdn',
    '.md',
    '.mkd',
    '.mdwn',
    '.mdtxt',
    '.mdtext',
    '.text',
    '.Rmd'
  ]) {
    if (path.endsWith(extension)) {
      return true;
    }
  }
  return false;
}
